print("import: 'samtranslator'")
import samtranslator

print("import: 'samtranslator.intrinsics'")
import samtranslator.intrinsics

print("import: 'samtranslator.model'")
import samtranslator.model

print("import: 'samtranslator.parser'")
import samtranslator.parser

print("import: 'samtranslator.plugins'")
import samtranslator.plugins

print("import: 'samtranslator.policy_template_processor'")
import samtranslator.policy_template_processor

print("import: 'samtranslator.public'")
import samtranslator.public

print("import: 'samtranslator.sdk'")
import samtranslator.sdk

print("import: 'samtranslator.swagger'")
import samtranslator.swagger

print("import: 'samtranslator.translator'")
import samtranslator.translator

print("import: 'samtranslator.validator'")
import samtranslator.validator

